

import pandas as pd
from abc import ABC, abstractmethod

# ---------------------------------------------------------

def read_worksheet(file, wks):
	"""read an excel worksheet 


	Args:
		file (str): excel file name
		wks (str): worksheet name

	Raises:
		Exception: Raised for file read fail, or worksheet missing

	Returns:
		DataFrame: Pandas DataFrame Object
	"""	
	try:
		df = pd.read_excel(file, sheet_name=wks).fillna("")
	except:
		raise Exception(f"CRITICAL: {file} File Read failed. Expected Worksheet named '{wks}' missing")
	return df

def read_excel(file):
	"""read excel file, all worksheet

	Args:
		file (str): excel file name

	Returns:
		dict: dictionary of dataframes.
	"""	
	dfd = pd.read_excel(file, None)
	for k,v in dfd.items():
		dfd[k] = v.fillna("")
	return dfd

def to_int(item):
	"""try to converts item to integer

	Args:
		item (str): input string

	Returns:
		int, str: returns integer value if it is number, else same as input string
	"""	
	try:
		return int(item)
	except:
		return item


class DeviceDetails():
	""" Device details operations

	Args:
		device_file (str): Excel device database file 

	Raises:
		Exception: Raised for input error: if device filename missing
		Exception: Raised for input error: if provided device file missing or read fails.

	Returns:
		DeviceDetails: DeviceDetails object
	"""	

	def __init__(self, device_file):
		"""instance initializer
		"""		
		self.device_filename = device_file
		self.verify_input_dev_file()

	def verify_input_dev_file(self):
		"""check if provided input device file 

		Raises:
			Exception: Raised for input error: if device filename missing
			Exception: Raised for input error: if provided device file missing or read fails.
		"""		
		if not self.device_filename:
			raise Exception(f'input error: input device file name {self.device_filename}')
		try:
			self.dev_details = self.read_device()								## require		
		except Exception as e:
			raise Exception(f'input error: input device file either missing or missing with necessary sheets\n{e}')

	def read_device(self):
		"""reads device database

		Returns:
			dict: dictionary of dataframe
		"""		
		dr = {'var':{}, 'table':None}
		dr['var']['device'] = read_worksheet(self.device_filename, 'var') 		
		dr['table'] = self.merged_table_columns()
		self.device_details = dr['var']['device']
		return dr

	def merged_table_columns(self):
		"""merges all different type of interfaces/protocols details in to a single dataframe.

		Returns:
			DataFrame: Pandas DataFrame object collecting all interfaces/protocols details 
		"""		
		dfd = read_excel(self.device_filename)
		del(dfd['var'])
		df_table = pd.DataFrame({'key':[]})
		for k, df in dfd.items():
			df_table = pd.concat([df_table, df], ignore_index=True).fillna("")
		for c in df_table:
			try:		
				df_table[c] = df_table[c].apply(to_int)
				df_table[c] = df_table[c].apply(str)
			except:
				pass
		return df_table

	def merge_with_var_frames(self, regional_frames):
		"""merge device var details with provided custom regional DataFrame(s) 
		custom regional frame variables/values  overrides device var variables/values.

		Args:
			regional_frames (list): list of custom regional DataFrames to be added to var.
		"""		
		frames = [self.device_details,]
		if regional_frames:
			frames.extend(regional_frames)
		self.var = self.merge_vars(frames)

	def read_table(self):
		"""reads table dataframe and add var/table to dataframe dictionary
		"""
		self.table = self.dev_details['table'].T.to_dict()
		self.data = {'var': self.var, 'table': self.table}

	def merge_vars(self, frames):
		"""merges var details from two different dataframes ( region and device - databases )
		(support definition)

		Args:
			frames (list): list of DataFrame(s) to be merged

		Returns:
			DataFrame: merged DataFrame (`var`)
		"""		
		return pd.concat(frames, ignore_index=True).set_index('var').to_dict()['default']



# --------------------------------------------------------------------------------------------------------
# Regional DataFrame input template class
# --------------------------------------------------------------------------------------------------------

class ABSRegion(ABC):
	"""Abstract Base Class Template to define custom/regional dataframe

	Args:
		device_details (DataFrame): Pandas DataFrame with device `var` information
		custom_data_file (str): custom datafile.

	Inherits:
		ABC (ABC): abstract base class

	Abstract Properties:
		frames(list) : must be defined in custom class method, which should return a list of DataFrame(s) to override `var` attributes. 


	"""	

	def __init__(self, device_details, custom_data_file):
		"""instance initializer

		"""		
		self.device_details = device_details
		self.custom_data_file = custom_data_file

	@property
	@abstractmethod
	def frames(self):
		"""must be defined in custom class method, which should return a list of DataFrame(s) to override `var` attributes. 
		"""		
		pass





import ipaddress as ip
from nettoolkit_common import LST
from nettoolkit.addressing import addressing as ip_addressing
from nettoolkit.addressing import IPv4
from nettoolkit.addressing import get_summaries as nt_get_summaries


def str_to_list(item):
	"""splits string and returns list, items separated by either `comma`, `enter` 

	Args:
		item (str, int, float): input string or number value(s)

	Returns:
		list: separated list of input items
	"""	
	if isinstance(item, (str, int, float) ):
		items= []
		csv = item.strip().split(",")
		for _ in csv:
			lsv = _.strip().split("\n")
			for i in lsv:
				items.append(i)
		return items
	else:
		return item

def space_separated(items):
	"""joins provided items (iterables) by `spaces`

	Args:
		items (list): input items

	Returns:
		str: joins items by space
	"""	
	return " ".join(items)

def comma_separated(items):
	"""joins provided items (iterables) by `comma`

	Args:
		items (list): input items

	Returns:
		str: joins items by comma
	"""
	return ",".join(items)

def list_append(lst, item):
	"""append an item to list

	Args:
		lst (list): input list
		item (str, number): item to be appeneded to list

	Returns:
		list: updated list
	"""	
	return lst.append(item)

def list_extend(lst, item):
	"""Extend the list of items to list

	Args:
		lst (list): input list
		item (list): list of items to be extended to input list

	Returns:
		list: updated list
	"""
	return lst.extend(item)

def list_sorted(lst):
	"""provided sorted elements in list

	Args:
		lst (list): input list

	Returns:
		list: updated list
	"""	
	return sorted(lst)

def convert_to_int(lst):
	"""convert numeric string type elements to integer type in a list.

	Args:
		lst (list): converts numeric eliments as integer

	Returns:
		list: updated list
	"""	
	return [ int(x) for x in lst]

def groups_of_nine(lst):
	"""breaks down provided list in to multiple groups with max. nine elements in each group

	Args:
		lst (list): input list

	Returns:
		list: updated list of (lists: containing 9 elements)
	"""	
	lst = LST.convert_vlans_list_to_range_of_vlans_list(lst)
	lst = [ str(_) for _ in lst ]
	return LST.split(lst, 9)	

def physical_if_allowed(vlan, table):
	"""condition: checks for `filter==physical` and `vlan in vlan_members`

	Args:
		vlan (str, int): vlan number
		table (dict): dataframe dictionary

	Returns:
		int: interface value of matching row
	"""	
	for key, data in table.items():
		if data['filter'].lower()=='physical' and int(vlan) in LST.expand_vlan_list(str_to_list(data['vlan_members'])):
			return data['interface']
	return ""

def remove_trailing_zeros(net):
	"""removes the trailing zeros from given ipv6 address

	Args:
		net (str): input ipv6 address

	Returns:
		str: updated ipv6 address by removing trailing zeros
	"""	
	while True:
		trimmers = ( "::0", ":0", "::")
		exit = True
		for t in trimmers:
			if net.endswith(t):
				net = net[:-1*len(t)]
				exit = False
		if exit: break
	return net


def ipv6_urpf_acl_network(subnet):
	"""provides ipv6 address network ip 

	Args:
		subnet (str): ipv6 address

	Returns:
		str: network address value for provided ipv6 address
	"""	
	pfx = ip.ip_interface(subnet)
	return str(pfx.network.network_address)

def nth_ip(net, n, withMask=False):
	"""get n-th ip address of given network.
	withMask: will return value along with mask otherwise only subnet 

	Args:
		net (str): input ipv4 address
		n (int): number (number of ip address to be return from subnet)
		withMask (bool, optional): return with mask or without mask. Defaults to False.

	Returns:
		str: nth ip address from subnet
	"""	
	try:
		_net = str(ip.ip_interface(net).network)
		v4 = ip_addressing(_net)
		return v4.n_thIP(n, True) if withMask else v4[n]
	except ValueError:
		return ""

def mask(net):
	"""get the subnet mask for given network (eg: 24)

	Args:
		net (str): input ipv4 address

	Returns:
		str: subnet mask
	"""	
	if net:
		_net = str(ip.ip_interface(net).network)
		v4 = ip_addressing(_net)
		return v4.mask
	else:
		return 'n/a'

def netmask(net):
	"""get network mask for given network (eg: 255.255.255.0)

	Args:
		net (str): input ipv4 address

	Returns:
		str: subnet mask
	"""	
	try:
		return str(ip.ip_interface(net).netmask)
	except:
		return ""

def invmask(net):
	"""get inverse mask for given network (eg: 0.0.0.255)

	Args:
		net (str): input ipv4 address

	Returns:
		str: subnet mask
	"""	
	v4 = v4addressing(net)
	return str(v4.invmask)

def addressing(net): 
	"""get the ip of given subnet

	Args:
		net (str): input ipv4 address

	Returns:
		str: ip address
	"""	
	return ip.ip_interface(net)

def int_to_str(data):
	"""get the actual physical interface value by removing training sub interfaces.

	Args:
		data (str): input interface string

	Returns:
		str: trunkated interface (after removal of sub-interface value)
	"""	
	return str(data).split(".")[0]

def v4addressing(ip, mask="32"):
	"""get the IPv4 objetct for given ip/mask (default mask=32)

	Args:
		ip (str): input ip address/mask value
		mask (str, optional): ip_mask. Defaults to "32".

	Returns:
		IPv4: IPv4 object
	"""	
	if ip.find("/") > 0: return IPv4(ip)
	return IPv4(ip+"/"+str(mask))

def get_summaries(lst_of_pfxs):
	"""get the summaries for provided prefixes.

	Args:
		lst_of_pfxs (list): list of prefixes to be summarized

	Returns:
		list: list of summarized prefixes
	"""	
	lst_of_pfxs = LST.remove_empty_members(lst_of_pfxs)
	try:
		return nt_get_summaries(*lst_of_pfxs)
	except:
		print(f"ERROR RECEIVE SUMMARIES")# {lst_of_pfxs}")
		return []

def iprint(x): 
	"""i print function to be use withing jinja template for debug.

	Args:
		x (str): value to be print during junja process
	"""	
	print(x)

def get_item(lst, n):
	"""get the nth item from list

	Args:
		lst (list): list containing various items
		n (int): index of item to be retrived 

	Returns:
		str: n-th item from list
	"""
	try:
		if isinstance(lst, (list, tuple)):
			return lst[n]
		else:
			return lst
	except:
		return lst

def as_path_repeat(asn, times):
	if not asn: return ""
	asn = asn.strip() + " "
	asp_prep_string = asn*times
	return asp_prep_string.strip()



import jinja2
from inspect import getmembers, isfunction, isclass, isroutine

from .data_collect import DeviceDetails
from .cmn import common_fn as cmn
from . import func as func
from .func import Vrf, Vlan, Physical, Bgp, Aggregated, Loopback, Ospf, Static
from .general import *


class PrepareConfig():
	"""boiler plate code class for start configuration preparation

	Args:
		data_file (str): Excel database
		jtemplate_file (str): Jinja Template
		output_folder (str, optional): output path. Defaults to ".".
		regional_file (str, optional): custom static regional variable file. Defaults to None. (overrides device var)
		regional_class (class, optional): custom class returning frames to be merge with device var . Defaults to None.

	Raises:
		Exception: Raise for Custom class insertion
		Exception: Raise for Custom module insertion
	"""	

	# -----------------------------------------
	# IMPORT FILTERS FOR JINJA VARIABLES
	# -----------------------------------------
	filters = {}
	filters.update({'Vrf': Vrf, 'Bgp': Bgp,
		'Vlan': Vlan, 'Physical': Physical, 'Aggregated': Aggregated, 'Loopback': Loopback, 		
		'Ospf': Ospf, 'Static': Static,
	})
	filters.update(dict(getmembers(cmn, isfunction)))
	filters.update(dict(getmembers(Vrf, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Vlan, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Physical, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Bgp, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Aggregated, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Loopback, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Static, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Ospf, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(func, isfunction)))

	def __init__(self,
		data_file,
		jtemplate_file,
		output_folder=".",
		regional_file=None,
		regional_class=None,
		):
		"""Object Initializer
		"""		
		self.data_file = data_file
		self.jtemplate_file = jtemplate_file.replace("\\", '/')
		self.output_folder = output_folder
		self.regional_file = regional_file
		self.regional_class = regional_class

	def custom_class_add_to_filter(self, **kwargs):
		"""add custom classes and its methods as jinja filters. External callable.
		"""
		for filtername, _cls in kwargs.items():
			try:
				self.filters.update({filtername: _cls})
				pre = self.filters.keys()
				self.filters.update(dict(getmembers(_cls, lambda x:not(isroutine(x))))['__dict__'] )
				post = self.filters.keys()
			except Exception as e:
				raise Exception(f"Class Insertion Failed for filter {filtername}\n{e}")

	def custom_module_methods_add_to_filter(self, *modules):
		"""add custom methods from module(s) as jinja filters. External callable.
		"""
		for module in modules:
			try:
				self.filters.update(dict(getmembers(module, isfunction)))
			except Exception as e:
				raise Exception(f"Module Insertion Failed {module}\n{e}")

	def start(self):
		"""kick start generation
		"""
		# ## LOAD - DATA
		DD = DeviceDetails(self.data_file)
		frames = []
		try:
			RCD = self.regional_class(DD.device_details, self.regional_file)
			frames = RCD.frames
		except:
			pass
		DD.merge_with_var_frames(frames)
		DD.read_table()

		# ## LOAD - JINJA TEMPLATE AND ENVIRONMENT
		templateLoader = jinja2.FileSystemLoader(searchpath='')
		templateEnv = jinja2.Environment(loader=templateLoader, 
			extensions=['jinja2.ext.loopcontrols', 'jinja2.ext.do',])
		for key, value in self.filters.items():
			templateEnv.filters[key] = value

		# ## TEMPLATE FILE		
		template = templateEnv.get_template(self.jtemplate_file)
		outputText = template.render(DD.data)#, undefined=jinja2.StrictUndefined) # Enable undefined for strict variable check

		# ## WRITE OUT
		model, template_ver = get_model_template_version(self.jtemplate_file)
		op_file = f"{self.output_folder}/{DD.data['var']['hostname']}-{model}-{template_ver}-j2Gen.cfg"
		with open(op_file, 'w') as f:
			f.write(outputText)

